#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <unistd.h>
#include <pthread.h>
#include <assert.h>

#include "liburing.h"

#define MAX_CONNECTIONS     4096
#define BACKLOG             512
#define MAX_MESSAGE_LEN     2048
#define BUFFERS_COUNT       MAX_CONNECTIONS

enum {
    ACCEPT,
    READ,
    WRITE,
    PROV_BUF,
};

typedef struct conn_info {
    __u32 fd;
    __u16 type;
    __u16 bid;
} conn_info;

char bufs[BUFFERS_COUNT][MAX_MESSAGE_LEN] = {0};
int group_id = 1337;
int pipes[2];
int control_pipes[2];
struct io_uring ring;
char buffer[1024];

void consume() {
    struct io_uring_cqe *cqe;
    int ret = io_uring_wait_cqe(&ring, &cqe);
    printf("consuming..\n");
    if (ret < 0) {
        printf("Error waiting for completion: %s\n",
                strerror(-ret));
        return;
    }
    if (!cqe) return;
    /* Now that we have the CQE, let's process the data */
    if (cqe->res < 0) {
        printf("Error in async operation: %s\n", strerror(-cqe->res));
        ;
    }
    printf("Result of the operation: %d\n", cqe->res);
    printf("cqe user_data: %lld\n", cqe->user_data);
    io_uring_cqe_seen(&ring, cqe);
}

void * work(void *) {
    struct io_uring_sqe * sqe;
    int ret;
    sqe = io_uring_get_sqe(&ring);
    io_uring_prep_read(sqe, control_pipes[0], buffer, 5, 0);
    sqe->flags |= IOSQE_IO_LINK;
    sqe->user_data = 1;


    sqe = io_uring_get_sqe(&ring);
    io_uring_prep_read(sqe, 0, buffer, 5, 0); // read from register file
    sqe->flags |= IOSQE_FIXED_FILE | IOSQE_IO_LINK;
    sqe->user_data = 2;

    io_uring_submit(&ring);
    return NULL;
}

int main(int argc, char *argv[]) {
    struct io_uring_params params;
    memset(&params, 0, sizeof(params));
    params.flags |= IORING_SETUP_SQPOLL;
    //params.sq_thread_idle = 3000;

    if (io_uring_queue_init_params(128, &ring, &params) < 0) {
        perror("io_uring_init_failed...\n");
        exit(1);
    }

    if (!(params.features & IORING_FEAT_FAST_POLL)) {
        printf("IORING_FEAT_FAST_POLL not available in the kernel, quiting...\n");
        exit(0);
    }

    struct io_uring_probe *probe;
    probe = io_uring_get_probe_ring(&ring);
    if (!probe || !io_uring_opcode_supported(probe, IORING_OP_PROVIDE_BUFFERS)) {
        printf("Buffer select not supported, skipping...\n");
        exit(0);
    }
    free(probe);
    struct io_uring_sqe *sqe;
    struct io_uring_cqe *cqe;

    sqe = io_uring_get_sqe(&ring);
    io_uring_prep_provide_buffers(sqe, bufs, MAX_MESSAGE_LEN, BUFFERS_COUNT, group_id, 0);
    io_uring_submit(&ring);
    io_uring_wait_cqe(&ring, &cqe);
    if (cqe->res < 0) {
        printf("cqe->res = %d\n", cqe->res);
        exit(1);
    }
    io_uring_cqe_seen(&ring, cqe);

    int ret;
    ret = pipe(control_pipes);
    assert(ret == 0);

    ret = pipe(pipes);
    assert(ret == 0);

    pthread_t thread;
    pthread_create(&thread, NULL, work, NULL);

    ret = io_uring_register_files(&ring, &pipes[0], 1);
    assert(ret == 0);
    io_uring_unregister_files(&ring);

    write(pipes[1], "abcde", 5);
    close(pipes[0]);
    write(control_pipes[1], "12345", 5);
    consume();
    consume();
    //close(pipes[1]);

	return 0;
}
